// FileData.cpp: implementation of the CFileData class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Filedata.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CFileData::CFileData() :
	m_lpszFileName(NULL), m_pStream(NULL), m_bInput(TRUE)
{
}

CFileData::~CFileData()
{
	Close();
}

CFileData::CFileData(LPCTSTR lpszFileName, BOOL bInput) :
	m_lpszFileName(NULL), m_pStream(NULL), m_bInput(TRUE)
{
	if (lpszFileName != NULL)
	{
		m_bInput = bInput;
		m_lpszFileName = lpszFileName;
		Open();
	}
}

BOOL CFileData::Flush()
{
	if (m_pStream != NULL)
	{
		return (fflush(m_pStream) != EOF); 
	}
	return FALSE;
}

BOOL CFileData::IsOpen()
{
	return(m_pStream != NULL);
}

void CFileData::Close()
{
	if (m_pStream != NULL)
	{
		fflush(m_pStream); 
		fclose(m_pStream);
		m_pStream = NULL;
	}
}

BOOL CFileData::Open(BOOL bAppend)
{
	if (m_lpszFileName == NULL) return FALSE;
	if (m_pStream != NULL) return TRUE;
#if _SECURE_ATL
	errno_t err;
	if (m_bInput)
		err = fopen_s(&m_pStream, m_lpszFileName, _T("r+b"));
	else
	{
		if (bAppend)
			err = fopen_s(&m_pStream, m_lpszFileName, _T("a+b"));
		else
			err = fopen_s(&m_pStream, m_lpszFileName, _T("w+b"));
	}
	_ASSERTE(err == NOERROR);
#else
	if (m_bInput)
		m_pStream = _tfopen(m_lpszFileName, _T("r+b"));
	else
	{
		if (bAppend)
			m_pStream = _tfopen(m_lpszFileName, _T("a+b"));
		else
			m_pStream = _tfopen(m_lpszFileName, _T("w+b"));
	}
#endif
	return IsOpen();
}

BOOL CFileData::Read(LPVOID lpszBuff, DWORD &dwLen, fpos_t pos)
{
	if (m_pStream != NULL && lpszBuff != NULL && dwLen > 0)
	{
		fpos_t fpos;
		if(fgetpos(m_pStream, &fpos) != 0 )
			return FALSE;
		if (fpos != pos)
			if(fsetpos(m_pStream, &pos) != 0 )
				return FALSE;

		size_t nLen = fread(lpszBuff, sizeof(char), (size_t)dwLen, m_pStream);
		dwLen = nLen;
		return (nLen > 0); 
	}
	return FALSE;
}

BOOL CFileData::Write(LPVOID lpszBuff, DWORD &dwLen, fpos_t pos)
{
	if (m_pStream != NULL && lpszBuff != NULL && dwLen > 0)
	{
		if (pos >= 0)
		{
			// Need set pos berfore
			fpos_t fpos;
			if(fgetpos(m_pStream, &fpos) != 0 )
				return FALSE;
			if (fpos != pos)
				if(fsetpos(m_pStream, &pos) != 0 )
					return FALSE;
		}
		
		size_t nLen = fwrite(lpszBuff, sizeof(_TCHAR), (size_t)dwLen, m_pStream);
		dwLen = nLen;
		return (nLen > 0); 
	}
	return FALSE;
}

// Get file length
DWORD CFileData::GetLength()
{
	if (m_pStream != NULL)
	{
		DWORD dwLength = 0;
		LONG nPos = ftell(m_pStream);
		if (fseek(m_pStream, 0L, SEEK_END) == 0)
		{
			dwLength = ftell(m_pStream);
			if (fseek(m_pStream, nPos, SEEK_SET) == 0)
				return dwLength;
			else
				return 0;
		}
	}
	return 0;
}
