// RegKeyTools.cpp: implementation of the RegKeyTools class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ShowErrors.h"
#include "RegFind.h"
#include "Filedata.h"
#include "RegKeyTools.h"
#include <winreg.h>
#include <time.h>
//#include <winsock2.h>

//////////////////////////////////////////////////////////////////////
// Static properties
//////////////////////////////////////////////////////////////////////

ROOTKEY CRegKeyTools::m_RootKeys[] = 
{
	HKEY_CLASSES_ROOT, _T("HKEY_CLASSES_ROOT"), 
	HKEY_CURRENT_USER, _T("HKEY_CURRENT_USER"), 
	HKEY_LOCAL_MACHINE, _T("HKEY_LOCAL_MACHINE"), 
	HKEY_USERS, _T("HKEY_USERS")
};

LPCTSTR CRegKeyTools::m_FileHeader[] = 
{
	_T("REGEDIT4\r\n\r\n"),
	_T(";    Created by %s v%s from computer: %s\r\n"),
	_T(";    on %s\r\n\r\n"),
	_T(";    Options: Keys - [%s], Values - [%s], Data - [%s], Whole strings only - [%s], Case sensitive - [%s].\r\n"),
	_T(";    for expression: %s\r\n\r\n")
};

WSADATA CRegKeyTools::m_wsaData;
_TCHAR CRegKeyTools::m_szFileIni[MAX_PATH];

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CRegKeyTools::CRegKeyTools() : 
	m_lpszFileIni(FILEINI),
	m_lpszSectionKeys(SECTIONKYES),
	m_bKeys(FALSE),
	m_bValues(TRUE),
	m_bData(FALSE),
	m_bWholeString(FALSE),
	m_bMatchCase(FALSE),
	m_bShowResult(FALSE),
	m_lpResultKeys(NULL),
	m_nMaxRootItems(0),
	m_hTerminate(NULL),
	m_bTerminated(FALSE),
	m_curFilePos(0),
	m_bIsModified(FALSE),
	m_bIsNewExcludeKeys(FALSE)
{
	m_nMaxRootItems = (sizeof(m_RootKeys) / sizeof(ROOTKEY));
	
	m_lpResultKeys = (LPBOOL)::new BOOL[m_nMaxRootItems];
	ClearResults();

	m_szCurKey[0] = (_TCHAR)'\0';
	m_szFileIni[0] = (_TCHAR)'\0'; 
	m_szFileOut[0] = (_TCHAR)'\0'; 
	m_szUserSID[0] = (_TCHAR)'\0'; 
	m_szSearchText[0] = (_TCHAR)'\0';
	m_szLocalHostName[0] = (CHAR)'\0';
	m_fileData.m_bInput = FALSE;

	BOOL bIsSettings = ReadSettings();
	_ASSERTE(bIsSettings);
	
	memset(&m_SearhInfo, 0, sizeof(SEARCHINFO));
	memset(&m_wsaData, 0, sizeof(WSADATA));
}

CRegKeyTools::~CRegKeyTools()
{
	if (m_lpResultKeys != NULL)
	{
		::delete [] m_lpResultKeys;
		m_lpResultKeys = NULL;
	}
	WriteSettings();
}

//////////////////////////////////////////////////////////////////////
// Member functions
//////////////////////////////////////////////////////////////////////

void CRegKeyTools::CloseFileData()
{
	if (m_fileData.IsOpen())
		m_fileData.Close(); 
	m_fileData.m_lpszFileName = NULL;
}

BOOL CRegKeyTools::GetFileIni()
{
	BOOL bRetVal = FALSE;
	if (_tcslen(m_szFileIni) == 0)
	{
		DWORD dwLen = ::GetModuleFileName(_Module.GetModuleInstance(), m_szFileIni, MAX_PATH);
		_ASSERTE(dwLen != 0);
		if (dwLen == 0)
		{
			ShowErrorMsg((HRESULT)GetLastError(), _T("CRegKeyTools::ReadSettings() in GetModuleFileName"));
			return bRetVal;
		}
		LPTSTR lpchPoint = _tcsrchr(m_szFileIni, (int)'.');
		_ASSERTE(lpchPoint != NULL);
		*(lpchPoint+1) = (_TCHAR)'\0';
#if _SECURE_ATL
		errno_t err = strcat_s(lpchPoint, _tcslen((LPCTSTR)m_szFileIni), _T("ini"));
		_ASSERTE(err == NOERROR);
#else
		_tcscat(lpchPoint, _T("ini"));
#endif
//		_tcscat(m_szFileIni, _T("\\"));
//		_tcscat(m_szFileIni, m_lpszFileIni);
		
		// Check ini-file exist & create if need
		if (::GetFileAttributes(m_szFileIni) == 0xFFFFFFFF)
		{
			// Close prev opened if need
			CloseFileData();
			m_curFilePos = 0;
			m_fileData.m_lpszFileName = m_szFileIni;

			// Create new ini-file
			BOOL bIsCreate = m_fileData.Open();
			_ASSERTE(bIsCreate);
			m_curFilePos = 0;
			
			// Fill section [ExcludeKeys]
			WriteStr(_T("["));
			WriteStr(m_lpszSectionKeys);
			WriteStr(_T("]\r\n\r\n"));

			// Close created ini-file
			CloseFileData();
			m_curFilePos = 0;
		}
		bRetVal = TRUE;
	}
	else
		bRetVal = TRUE;
	return bRetVal;
}

BOOL CRegKeyTools::ReadSettings()
{
	BOOL bRetVal = FALSE;
	if (!GetFileIni())
		return bRetVal;

	DWORD dwBuffLen = MAX_PATH*MAXEXCLUDEKEYS+2;
	_TCHAR szReturnedStrings[MAX_PATH*MAXEXCLUDEKEYS+2];
	DWORD dwLen = GetPrivateProfileSection(
		m_lpszSectionKeys,	// section name
		szReturnedStrings,  // return buffer
		dwBuffLen,			// size of return buffer
		m_szFileIni			// initialization file name
	);

	if (dwLen != 0)
	{
		LPTSTR lpszCutPos = szReturnedStrings;
		size_t nCurLen = 0;
		USES_CONVERSION;
		while (dwLen > 0)
		{
			nCurLen = _tcslen(lpszCutPos);
			if (nCurLen > 0)
				m_vecExcludeKeys.push_back(T2A(lpszCutPos)); 
			dwLen -= ++nCurLen;
			lpszCutPos += nCurLen;
		}
//		if (_tcslen(m_szUserSID) != 0)
//			m_vecExcludeKeys.push_back(T2A(m_szUserSID)); 
		bRetVal = TRUE;
	}
	else
	{
/*
		_TCHAR szMsg[MAX_PATH*2];
		_tcscpy(szMsg, _T("Can not found file: "));
		_tcscat(szMsg, m_szFileIni);
		_tcscat(szMsg, _T(" or not found section: "));
		_tcscat(szMsg, m_lpszSectionKeys);
		::MessageBox(NULL, szMsg, _T("Warning:"), MB_OK|MB_ICONWARNING);
*/
	}

	// m_bKeys
	dwLen = GetPrivateProfileString(
		SECTIONOPTS,        // section name
		OPTKYES,			// key name
		OPTKYESDEF,			// default string
		szReturnedStrings,  // destination buffer
		dwBuffLen,          // size of destination buffer
		m_szFileIni			// initialization file name
	);
	if (dwLen)
	{
#if _SECURE_ATL
		m_bKeys = StrToBool(szReturnedStrings, dwBuffLen);
#else
		m_bKeys = StrToBool(szReturnedStrings);
#endif
	}

	// m_bValues
	dwLen = GetPrivateProfileString(
		SECTIONOPTS,        // section name
		OPTVALUES,			// key name
		OPTVALUESDEF,		// default string
		szReturnedStrings,  // destination buffer
		dwBuffLen,          // size of destination buffer
		m_szFileIni			// initialization file name
	);
	if (dwLen)
	{
#if _SECURE_ATL
		m_bValues = StrToBool(szReturnedStrings, dwBuffLen);
#else
		m_bValues = StrToBool(szReturnedStrings);
#endif
	}

	// m_bData
	dwLen = GetPrivateProfileString(
		SECTIONOPTS,        // section name
		OPTDATA,			// key name
		OPTDATADEF,			// default string
		szReturnedStrings,  // destination buffer
		dwBuffLen,          // size of destination buffer
		m_szFileIni			// initialization file name
	);
	if (dwLen)
	{
#if _SECURE_ATL
		m_bData = StrToBool(szReturnedStrings, dwBuffLen);
#else
		m_bData = StrToBool(szReturnedStrings);
#endif
	}

	// m_bWholeString
	dwLen = GetPrivateProfileString(
		SECTIONOPTS,        // section name
		OPTWHOLESTR,		// key name
		OPTWHOLESTRDEF,		// default string
		szReturnedStrings,  // destination buffer
		dwBuffLen,          // size of destination buffer
		m_szFileIni			// initialization file name
	);
	if (dwLen)
	{
#if _SECURE_ATL
		m_bWholeString = StrToBool(szReturnedStrings, dwBuffLen);
#else
		m_bWholeString = StrToBool(szReturnedStrings);
#endif
	}

	// m_bMatchCase
	dwLen = GetPrivateProfileString(
		SECTIONOPTS,        // section name
		OPTMATCHCASE,		// key name
		OPTMATCHCASEDEF,	// default string
		szReturnedStrings,  // destination buffer
		dwBuffLen,          // size of destination buffer
		m_szFileIni			// initialization file name
	);
	if (dwLen)
	{
#if _SECURE_ATL
		m_bMatchCase = StrToBool(szReturnedStrings, dwBuffLen);
#else
		m_bMatchCase = StrToBool(szReturnedStrings);
#endif
	}

	bRetVal = TRUE;
	return bRetVal;
}

void CRegKeyTools::WriteSettings()
{
	if (!m_bIsModified)
		return ;

	if (!GetFileIni())
		return ;

	// m_bKeys
	WritePrivateProfileString(
		SECTIONOPTS,		// section name
		OPTKYES,			// key name
		BoolToStr(m_bKeys),	// string to add
		m_szFileIni			// initialization file
	);

	// m_bValues
	WritePrivateProfileString(
		SECTIONOPTS,			// section name
		OPTVALUES,				// key name
		BoolToStr(m_bValues),	// string to add
		m_szFileIni				// initialization file
	);

	// m_bData
	WritePrivateProfileString(
		SECTIONOPTS,		// section name
		OPTDATA,			// key name
		BoolToStr(m_bData),	// string to add
		m_szFileIni			// initialization file
	);

	// m_bWholeString
	WritePrivateProfileString(
		SECTIONOPTS,				// section name
		OPTWHOLESTR,				// key name
		BoolToStr(m_bWholeString),	// string to add
		m_szFileIni					// initialization file
	);

	// m_bMatchCase
	WritePrivateProfileString(
		SECTIONOPTS,				// section name
		OPTMATCHCASE,				// key name
		BoolToStr(m_bMatchCase),	// string to add
		m_szFileIni					// initialization file
	);
}

LPCTSTR CRegKeyTools::BoolToStr(BOOL bOption)
{
	return (bOption ? _T("On") : _T("Off"));
}

#if _SECURE_ATL
BOOL CRegKeyTools::StrToBool(LPTSTR lpszOpt, size_t lenBuff)
#else
BOOL CRegKeyTools::StrToBool(LPTSTR lpszOpt)
#endif
{
	BOOL bRetVal = FALSE;
#if _SECURE_ATL
	errno_t err = _strupr_s(lpszOpt, lenBuff);
	_ASSERTE(err == NOERROR);
#else
	_strupr(lpszOpt);
#endif
	if (!(bRetVal = (_tcsstr(lpszOpt, _T("ON")) != NULL)))
		bRetVal = (_tcsstr(lpszOpt, _T("YES")) != NULL);
	return bRetVal;
}

void CRegKeyTools::ClearResults()
{
	if (m_lpResultKeys != NULL)
	{
		BOOL* lpCurVal = m_lpResultKeys;
		for (int nItem = 0; nItem < m_nMaxRootItems; nItem++)
			(*lpCurVal++) = FALSE;
	}
}

BOOL CRegKeyTools::Find()
{
	BOOL bRetVal = TRUE;
	m_bTerminated = FALSE;
	LPSTATE lpCurState = m_SearhInfo.lpCurState;
	HWND hWndNotify = m_SearhInfo.hWndNotify; 
	
	// Clear state information
	if (lpCurState != NULL)
	{
		memset(lpCurState, 0, sizeof(STATE));
		if (hWndNotify != NULL)
		{
			::PostMessage(hWndNotify, UM_STATECHANGED, 0, 0);
			::Sleep(0);
		}
	}

	// Check file-out - close if need
	CloseFileData();
	m_szFileOut[0] = (_TCHAR)'\0';
	m_curFilePos = 0;

	if (m_lpResultKeys != NULL)
	{
		ClearResults();
		BOOL* lpCurVal = m_lpResultKeys;
		LPROOTKEY lpCurKey = m_RootKeys;
		
#if _SECURE_ATL
		errno_t err = strcpy_s(m_szSearchTxt, MAX_PATH, m_szSearchText);
		_ASSERTE(err == NOERROR);
		if (!m_bMatchCase)
		{
			err = _strupr_s(m_szSearchTxt, MAX_PATH);
			_ASSERTE(err == NOERROR);
		}
#else
		_tcscpy(m_szSearchTxt, m_szSearchText);
		if (!m_bMatchCase)
			_tcsupr(m_szSearchTxt);
#endif

		// For each top-root element
		for (int nItem = 0; nItem < m_nMaxRootItems; nItem++)
		{
			// Prepare current root-key
#if _SECURE_ATL
			err = strcpy_s(m_szCurKey, MAXKEYLEN, (*lpCurKey).lpszName);
			_ASSERTE(err == NOERROR);
#else
			_tcscpy(m_szCurKey, (*lpCurKey).lpszName);
#endif
			// Search for current item & check result
			if (!((*lpCurVal++) = FindInNode((*lpCurKey++).hKey)))
				bRetVal = FALSE;

			// Notify message about changed state
			if (lpCurState != NULL && hWndNotify != NULL)
			{
				::PostMessage(hWndNotify, UM_STATECHANGED, 0, 0);
				::Sleep(0);
			}

			// Check terminate search-process
			if (m_bTerminated)
			{
				bRetVal = FALSE;
				break;
			}
		}

		// Clear current root-key
		m_szCurKey[0] = (_TCHAR)'\0';
	}
	else
	{
		// m_lpResultKeys == NULL 
		bRetVal = FALSE;
	}

	if (m_bTerminated && m_fileData.IsOpen())
	{
		// Write message - search terminated
		Write(SEARCHTERMINATED);
	}

	CloseFileData();
//	m_curFilePos = 0;

	if (lpCurState->dwFoundItems == 0)
	{
		// File-results is empty -> delete
		if (::GetFileAttributes(m_szFileOut) != 0xFFFFFFFF)
		{
			if (::DeleteFile(m_szFileOut))
				m_szFileOut[0] = (_TCHAR)'\0';
		}
		else
			m_szFileOut[0] = (_TCHAR)'\0';
	}

	if (m_bIsNewExcludeKeys && m_vecExcludeKeys.size() > 0)
	{
		BOOL bIsSID = FALSE;
/*
		if (_tcslen(m_szUserSID) != 0)
		{
			bIsSID = TRUE;
		}
*/

		m_fileData.m_lpszFileName = FILEKEYSEXCLUDE;
		if (m_fileData.Open())
		{
			USES_CONVERSION;
			LPCTSTR lpszCurKey = NULL;
			STRVECTOR::iterator Iter;
			m_curFilePos = 0;
			for (Iter = m_vecExcludeKeys.begin(); Iter != m_vecExcludeKeys.end(); Iter++)
			{
				lpszCurKey = A2CT((const char*)(*Iter).c_str());
				if (bIsSID && (_tcscmp(lpszCurKey, m_szUserSID) == 0))
					continue;
				WriteStr(lpszCurKey);
				WriteStr(_T("\r\n"));
			}
			m_fileData.Close();
			m_curFilePos = 0;
		}
		m_fileData.m_lpszFileName = NULL;
		m_bIsNewExcludeKeys = FALSE;
	}
	return bRetVal;
}

BOOL CRegKeyTools::WriteCurKey()
{
	return (Write(_T("[")) && Write(m_szCurKey) && Write(_T("]\r\n")));
}

BOOL CRegKeyTools::WriteValueName(LPCTSTR lpszValueName, DWORD dwValue)
{
	if (dwValue == 0)
		return Write(_T("@"));
	else
		return (Write(_T("\"")) && Write(lpszValueName) && Write(_T("\"")));
}

// Compare(lpszDataStr, m_szSearchTxt) 
#if _SECURE_ATL
	BOOL CRegKeyTools::CompareStr(LPCTSTR lpszDataStr, LPTSTR lpszBuffer, BOOL bInKey, size_t lenBuffer)
#else
	BOOL CRegKeyTools::CompareStr(LPCTSTR lpszDataStr, LPTSTR lpszBuffer, BOOL bInKey)
#endif
{
	BOOL bFound = FALSE;
	LPCTSTR lpszCurData = NULL;

	if (bInKey)
		lpszCurData = (lpszDataStr == NULL ? m_szCurKey : lpszDataStr);
	else
	{
		if (lpszDataStr == NULL)
			return bFound;
		lpszCurData = lpszDataStr;
	}

	if (!m_bMatchCase)
	{
		// Use lpszBuffer as temp storege for UpperCase
		// Need length(lpszBuffer) >= length(lpszDataStr) !!!
		_ASSERTE(lpszBuffer != NULL);
#if _SECURE_ATL
		errno_t err = strcpy_s(lpszBuffer, lenBuffer, lpszCurData);
		_ASSERTE(err == NOERROR);
		err = _tcsupr_s(lpszBuffer, lenBuffer);
		_ASSERTE(err == NOERROR);
#else
		_tcscpy(lpszBuffer, lpszCurData);
		_tcsupr(lpszBuffer);
#endif
		if (m_bWholeString)
			bFound = (_tcscmp(lpszBuffer, m_szSearchTxt) == 0);
		else
			bFound = (_tcsstr(lpszBuffer, m_szSearchTxt) != NULL);
	}
	else
	{
		if (m_bWholeString)
			bFound = (_tcscmp(lpszCurData, m_szSearchTxt) == 0);
		else
			bFound = (_tcsstr(lpszCurData, m_szSearchTxt) != NULL);
	}
	
	return bFound;
}

BOOL CRegKeyTools::FindInNode(HKEY hKey, LPTSTR lpszSubkey)
{
	// Check terminate search-process
	if (m_bTerminated)
		return FALSE;

	BOOL bRetVal = FALSE;
	BOOL bEnabledKey = TRUE;
	if (lpszSubkey != NULL)
	{
		// Add Subkey to current key
#if _SECURE_ATL
		errno_t err = strcat_s(m_szCurKey, MAXKEYLEN, _T("\\"));
		_ASSERTE(err == NOERROR);
		err = strcat_s(m_szCurKey, MAXKEYLEN, lpszSubkey);
		_ASSERTE(err == NOERROR);
#else
		_tcscat(m_szCurKey, _T("\\"));
		_tcscat(m_szCurKey, lpszSubkey);
#endif
		
		// Check exclude keys
		if (m_vecExcludeKeys.size() > 0) 
		{
			USES_CONVERSION;
			LPCSTR lpszCurSubKey = T2CA(lpszSubkey);
			STRVECTOR::iterator Iter;
			for (Iter = m_vecExcludeKeys.begin(); Iter != m_vecExcludeKeys.end(); Iter++)
			{
				if (_tcsicmp((const char*)(*Iter).c_str(), m_szCurKey) == 0)
				{
					// No search for exclude key
					bEnabledKey = FALSE;
					bRetVal = TRUE;
					break;
				}
			}
		}
		
	}

	if (bEnabledKey)
	{

		LPSTATE lpCurState = m_SearhInfo.lpCurState;
		BOOL bFound = FALSE;
		BOOL bStateChanged = FALSE;

		if (m_bKeys)
		{
#if _SECURE_ATL
			if (bFound = CompareStr(lpszSubkey, m_szBuffer, TRUE, MAX_PATH))
#else
			if (bFound = CompareStr(lpszSubkey, m_szBuffer, TRUE))
#endif
			{
				if (lpCurState != NULL)
				{
	//				lpCurState->CS_StateOptions.Lock();  
					lpCurState->dwFoundItems++;
	//				lpCurState->CS_StateOptions.Unlock(); 

					bStateChanged = TRUE;
					if (WriteCurKey())
					{
					}
					else
					{
						// Error on write
						_ASSERTE(FALSE);
					}
				}
			}
		}
		
		// Try open subkey
		HKEY hCurKey = NULL;
		LONG lnCode = RegOpenKeyEx(
			hKey,					// handle to open key
			lpszSubkey,				// subkey name
			NULL,					// reserved
			KEY_READ,				// security access mask
			&hCurKey				// handle to open key
		);

		if (lnCode == ERROR_SUCCESS && hCurKey != NULL)
		{
			DWORD cSubKeys = NULL;			// number of subkeys
			DWORD cMaxSubKeyLen = NULL;		// longest subkey name
			DWORD cValues = NULL;			// number of value entries
			DWORD cMaxValueNameLen = NULL;	// longest value name
			DWORD cMaxValueLen = NULL;		// longest value data
			
			// Increment count kes
			if (lpCurState != NULL)
			{
				lpCurState->dwKeys++;
				bStateChanged = TRUE;
			}
			
			// Query current subkey info
			lnCode = RegQueryInfoKey(
				hCurKey,				// handle to key
				NULL,					// class buffer
				NULL,					// size of class buffer
				NULL,					// reserved
				&cSubKeys,				// number of subkeys
				&cMaxSubKeyLen,			// longest subkey name
				NULL,					// longest class string
				&cValues,				// number of value entries
				&cMaxValueNameLen,		// longest value name
				&cMaxValueLen,			// longest value data
				NULL,					// descriptor length
				NULL					// last write time
			);
			
			if (lnCode == ERROR_SUCCESS)
			{
				// Current subkey values
				if (m_bValues && cValues > 0)
				{
					DWORD dwMaxValueNameLen = cMaxValueNameLen+1;
					LPTSTR lpszValueName = (LPTSTR)::new _TCHAR[(int)dwMaxValueNameLen];
					DWORD dwMaxValueLen = cMaxValueLen+1;
					LPBYTE lpBuffData = (LPBYTE)::new BYTE[(int)dwMaxValueLen];
					size_t lenBuff = (int)max(dwMaxValueNameLen, dwMaxValueLen);
					LPTSTR lpBuffStr = (LPTSTR)::new _TCHAR[lenBuff];
					if (lpBuffData != NULL && lpszValueName != NULL && lpBuffStr != NULL)
					{
						LONG lnCurCode = 0;
						DWORD dwValue = 0;
						DWORD dwType = 0;
						DWORD dwData = 0;
						for (DWORD dwIndex = 0; dwIndex < cValues; dwIndex++)
						{
							dwValue = dwMaxValueNameLen;
							dwData = dwMaxValueLen;
							lnCurCode = RegEnumValue(
								hCurKey,		// handle to key to query
								dwIndex,		// index of value to query
								lpszValueName,	// value buffer
								&dwValue,		// size of value buffer
								NULL,			// reserved
								&dwType,		// type buffer
								lpBuffData,		// data buffer
								&dwData			// size of data buffer
							);

							if (lnCurCode == ERROR_SUCCESS)
							{
								// Ok, is value
								switch (dwType)
								{
								case REG_BINARY :
								//	Binary data in any form.
									break;
								case REG_DWORD :
								//	A 32-bit number.
									break;
								case REG_LINK :
								//	A Unicode symbolic link
									break;
								case REG_MULTI_SZ :
								//	An array of null-terminated strings, terminated by two null characters.
									break;
								case REG_NONE :
								//	No defined value type.
									break;
								case REG_RESOURCE_LIST :
								//	A device-driver resource list.
									break;
								case REG_SZ :
								//	A null-terminated string. It will be a Unicode or ANSI string, depending on whether you use the Unicode or ANSI functions.
								case REG_EXPAND_SZ :
								//	A null-terminated string that contains unexpanded references to environment variables
									*(lpBuffData+dwData) = (_TCHAR)'\0';
#if _SECURE_ATL
									if (CompareStr((LPTSTR)lpBuffData, lpBuffStr, FALSE, lenBuff))
#else
									if (CompareStr((LPTSTR)lpBuffData, lpBuffStr))
#endif
									{
										if (!bFound)
											WriteCurKey();
										bFound = TRUE;
										if (lpCurState != NULL)
										{
											lpCurState->dwFoundItems++;
											bStateChanged = TRUE;
											if (WriteValueName(lpszValueName, dwValue) && Write(_T("="))
												&& Write(_T("\"")) && Write((LPTSTR)lpBuffData) && Write(_T("\"")) && Write(_T("\r\n")))
											{
											}
											else
											{
												// Error on write
												_ASSERTE(FALSE);
											}
										}
									}
									break;
								default: ;
	//							case REG_DWORD_LITTLE_ENDIAN :
	//							case REG_DWORD_BIG_ENDIAN :
	//							case REG_QWORD_LITTLE_ENDIAN :
	//							case REG_QWORD :
								}
 
								// Increment count values
								if (lpCurState != NULL)
								{
	//								::EnterCriticalSection(&lpCurState->CS_StateOptions); 
									lpCurState->dwValues++;
	//								::LeaveCriticalSection(&lpCurState->CS_StateOptions);
									bStateChanged = TRUE;
								}
							}
							else
							{
								// Error in RegEnumValue
								if (lpszSubkey != NULL)
								{
									m_vecExcludeKeys.push_back(T2A(m_szCurKey));
									m_bIsNewExcludeKeys = TRUE;
									break;
								}
							}
						}
					}
					if (lpBuffStr != NULL)
					{
						:: delete [] lpBuffStr;
						lpBuffData = NULL;
					}
					if (lpBuffData != NULL)
					{
						:: delete [] lpBuffData;
						lpBuffData = NULL;
					}
					if (lpszValueName != NULL)
					{
						:: delete [] lpszValueName;
						lpszValueName = NULL;
					}
					
				}
				else
				{
					// Ok no cSubKeys
					bRetVal = TRUE;
				}

				if (bFound)
				{
					if (!Write(_T("\r\n")))
					{
						// Error on write
						_ASSERTE(FALSE);
					}
				}

				bRetVal = FALSE;
				if (cSubKeys > 0 && cMaxSubKeyLen > 0)
				{
					// For each sub-subkyes for current subkey
					DWORD dwMaxSubKeyLen = cMaxSubKeyLen+1;
					LPTSTR lpszName = (LPTSTR)::new _TCHAR[(int)dwMaxSubKeyLen];
					if (lpszName != NULL)
					{
						LONG lnCurCode = 0;
						DWORD cName = 0;
						bRetVal = TRUE; // No errors
						for(DWORD dwInex = 0; dwInex < cSubKeys; dwInex++)
						{
							// Check terminate search-process
							if (m_bTerminated)
							{
								bRetVal = FALSE;
								break;
							}

							cName = dwMaxSubKeyLen;
							lnCurCode = RegEnumKeyEx(
								hCurKey,	// handle to key to enumerate
								dwInex,		// subkey index
								lpszName,   // subkey name
								&cName,     // size of subkey buffer
								NULL,		// reserved
								NULL,		// class string buffer
								NULL,		// size of class string buffer
								NULL		// last write time
							);
							
							// Check terminate search-process
							if (m_hTerminate != NULL)
							{
								if (WaitForSingleObject(m_hTerminate, 0) == WAIT_OBJECT_0)
								{
									::ResetEvent(m_hTerminate);
									m_bTerminated = TRUE;
									bRetVal = FALSE;
									break;
								}
							}

							if (lnCurCode == ERROR_SUCCESS)
							{
								// Recursive search in current sub-subkey
								if (!FindInNode(hCurKey, lpszName))
								{
									// Error in FindInRoot 
									bRetVal = FALSE;
								}
							}
							else
							{
								// Error in RegEnumKeyEx
								if (lpszSubkey != NULL)
								{
									m_vecExcludeKeys.push_back(T2A(m_szCurKey));
									m_bIsNewExcludeKeys = TRUE;
									break;
								}
							}
						}
						::delete [] lpszName;
						lpszName = NULL;
					}
					else
					{
						// Out of memory
						bRetVal = FALSE;
						ShowErrorMsg((HRESULT)GetLastError(), _T("Out of memory"));
					}
				}
				else
				{
					// Ok no cSubKeys
					bRetVal = TRUE;
				}
			}
			else
			{
				// Error in RegQueryInfoKey
				if (lpszSubkey != NULL)
				{
					m_vecExcludeKeys.push_back(T2A(m_szCurKey));
					m_bIsNewExcludeKeys = TRUE;
				}
				if (bFound)
				{
					if (!Write(_T("\r\n")))
					{
						// Error on write
						_ASSERTE(FALSE);
					}
				}
			}
			RegCloseKey(hCurKey);
		}
		else
		{
			// Error in RegOpenKeyEx
			if (lpszSubkey != NULL)
			{
				m_vecExcludeKeys.push_back(T2A(m_szCurKey));
				m_bIsNewExcludeKeys = TRUE;
			}
			if (bFound)
			{
				if (!Write(_T("\r\n")))
				{
					// Error on write
					_ASSERTE(FALSE);
				}
			}
		}
		// Notify message about changed state
		if (bStateChanged 
			&& m_SearhInfo.hWndNotify != NULL 
			&& (lpCurState->dwKeys % 100) == 0)
		{
			::PostMessage(m_SearhInfo.hWndNotify, UM_STATECHANGED, 0, 0);
			::Sleep(0);
		}
	}

	if (lpszSubkey != NULL)
	{
		// Remove last subkey from current key 
		LPTSTR lpLast = _tcsrchr(m_szCurKey, (int)'\\');
		if (lpLast != NULL)
			*lpLast = (_TCHAR)'\0';
	}
	return bRetVal;
}

BOOL CRegKeyTools::IsTerminated()
{
	return m_bTerminated;
}

BOOL CRegKeyTools::GetLocalHostName(LPSTR lpszLocalHostName, size_t &nlen)
{
   	BOOL bRetVal = FALSE;

	if (lpszLocalHostName == NULL)
		return bRetVal;

	if (nlen == 0)
		return bRetVal;

	size_t nLenName = strlen(m_szLocalHostName);
	if (nLenName == 0)
	{
		// Get LocalHostName
		WORD wVersionRequested = MAKEWORD(2, 2);
		int err = WSAStartup(wVersionRequested, &m_wsaData);
		if (err != S_OK)
		{
			ShowErrorMsg((HRESULT)err, _T(" in CRegKeyTools::GetLocalHostName [WSAStartup()0"));
			return FALSE;
		}
		
		if ((err = gethostname(m_szLocalHostName, MAXHOSTLEN)) != S_OK)
		{
			ShowErrorMsg((HRESULT)err, _T(" in CRegKeyTools::GetLocalHostName [gethostname()]"));
			WSACleanup();
			return FALSE;
		}
		nLenName = strlen(m_szLocalHostName);
		WSACleanup();

		if (!isupper((int)m_szLocalHostName[0]))
		{
			CHAR szFist[2];
			szFist[0] = m_szLocalHostName[0];
			szFist[1] = (CHAR)'\0';
#if _SECURE_ATL
			errno_t err = _strupr_s(szFist, 2);
			_ASSERTE(err == NOERROR);
#else
			_strupr(szFist);
#endif
			m_szLocalHostName[0] = szFist[0];
		}
	}

	bRetVal = (nLenName > 0);
	if (lpszLocalHostName == NULL)
	{
		nlen = nLenName;
		return bRetVal;
	}
	if (nLenName && nlen)
	{
#if _SECURE_ATL
		errno_t err;
		if (nlen > nLenName)
		{
			err = strcpy_s(lpszLocalHostName, nlen, m_szLocalHostName);
			_ASSERTE(err == NOERROR);
			nlen = nLenName;
		}
		else
		{
			err = strncpy_s(lpszLocalHostName, nlen, m_szLocalHostName, nlen);
			_ASSERTE(err == NOERROR);
		}
#else
		if (nlen > nLenName)
		{
			strcpy(lpszLocalHostName, m_szLocalHostName);
			nlen = nLenName;
		}
		else
			strncpy(lpszLocalHostName, m_szLocalHostName, nlen);
#endif
	}
	return bRetVal;
}

BOOL CRegKeyTools::CreateFile()
{
	BOOL bRetVal = FALSE;
	
	// Get LocalHostName
	CHAR szLocalHostName[MAXHOSTLEN];
	size_t nlen = MAXHOSTLEN;
	if (!GetLocalHostName(szLocalHostName, nlen))
		return bRetVal;

	// Get current datetime
	time(&m_sysTime);
#if _SECURE_ATL
	errno_t err = localtime_s(&m_curTime, &m_sysTime);
	_ASSERTE(err == NOERROR);
#else
	m_curTime = localtime(&m_sysTime);
#endif
	_tcsftime(m_szCurTime, MAXDATETIME, _T("%Y%m%d %H%M%S"), &m_curTime);

	// GetCurrentDirectory
	_TCHAR szCurDir[MAX_PATH];
	if (GetCurrentDirectory(MAX_PATH, szCurDir) == 0)
	{
		ShowErrorMsg((HRESULT)GetLastError(), _T(" in CRegKeyTools::CreateFile [GetCurrentDirectory()]"));
		return bRetVal;
	}
	
	// Get new file name
	USES_CONVERSION;
#if _SECURE_ATL
	sprintf_s(m_szFileOut, MAX_PATH, _T("%s\\%s %s %s%s"), szCurDir, UTILNAME,
		A2CT(szLocalHostName), m_szCurTime, REGFILEEXT);
#else
	_stprintf(m_szFileOut, _T("%s\\%s %s %s%s"), szCurDir, UTILNAME,
		A2CT(szLocalHostName), m_szCurTime, REGFILEEXT);
#endif

	// Try create
	m_fileData.m_lpszFileName = m_szFileOut;
	if (!m_fileData.Open())
	{
		ShowErrorMsg((HRESULT)GetLastError(), _T(" in CRegKeyTools::CreateFile [m_fileData.Open()]"));
		return bRetVal;
	}
	bRetVal = TRUE;
	return bRetVal;
}

LPCTSTR CRegKeyTools::GetFileResult()
{
	return m_szFileOut;
}

BOOL CRegKeyTools::WriteStr(LPCTSTR lpszData)
{
	DWORD dwLen = _tcslen(lpszData);
	if (m_fileData.Write((LPVOID)lpszData, dwLen))
	{
		m_curFilePos += dwLen;
		return TRUE;
	}
	else
		return FALSE;
}

BOOL CRegKeyTools::Write(LPCTSTR lpszData)
{
	BOOL bRetVal = FALSE;
	if(!m_fileData.IsOpen())
	{
		if (!CreateFile())
		{
			return bRetVal;
		}

		// Inset header in created file
		if (!(bRetVal = WriteStr(m_FileHeader[0])))
			return bRetVal;
			
		USES_CONVERSION;
		size_t nLenBuff = 2*MAX_PATH;
		_TCHAR szLine[2*MAX_PATH];
#if _SECURE_ATL
		sprintf_s(szLine, nLenBuff, m_FileHeader[1], UTILNAME, UTILVERS, A2CT(m_szLocalHostName));
#else
		_stprintf(szLine, m_FileHeader[1], UTILNAME, UTILVERS, A2CT(m_szLocalHostName));
#endif
		if (!(bRetVal = WriteStr(szLine)))
			return bRetVal;

		_tcsftime(m_szCurTime, MAXDATETIME, _T("%#c"), &m_curTime);
#if _SECURE_ATL
		sprintf_s(szLine, nLenBuff, m_FileHeader[2], m_szCurTime);
#else
		_stprintf(szLine, m_FileHeader[2], m_szCurTime);
#endif
		if (!(bRetVal = WriteStr(szLine)))
			return bRetVal;

#if _SECURE_ATL
		sprintf_s(szLine, nLenBuff, m_FileHeader[3], 
			BoolToStr(m_bKeys), BoolToStr(m_bValues), BoolToStr(m_bData),
			BoolToStr(m_bWholeString), BoolToStr(m_bMatchCase));
#else
		_stprintf(szLine, m_FileHeader[3], 
			BoolToStr(m_bKeys), BoolToStr(m_bValues), BoolToStr(m_bData),
			BoolToStr(m_bWholeString), BoolToStr(m_bMatchCase));
#endif
		if (!(bRetVal = WriteStr(szLine)))
			return bRetVal;

#if _SECURE_ATL
		sprintf_s(szLine, nLenBuff, m_FileHeader[4], m_szSearchText);
#else
		_stprintf(szLine, m_FileHeader[4], m_szSearchText);
#endif
		if (!(bRetVal = WriteStr(szLine)))
			return bRetVal;
	}

	bRetVal = WriteStr(lpszData);
	return bRetVal;
}

void CRegKeyTools::SetModified(BOOL bModified)
{
	m_bIsModified = bModified;
}

BOOL CRegKeyTools::GetUserSID()
{
	BOOL bRetVal = FALSE;
	DWORD dwError = 0;
	HANDLE hProcess = NULL;
	HANDLE hToken = NULL;
	DWORD dwTuBuff = 0;
	DWORD dwSIDLen = MAXSIDLEN;
	PTOKEN_USER pTuInfo = NULL;
	BOOL bIsSupport = FALSE;
	try
	{
		// GetCurrentProcess
		hProcess = ::GetCurrentProcess();
		
		// Obtain current process token.
		if (!::OpenProcessToken(hProcess, TOKEN_QUERY, &hToken))
		{
			dwError = ::GetLastError(); 
			throw _T("CRegKeyTools::GetUserSID [OpenProcessToken()]");
		}
		bIsSupport = TRUE;

		// Obtain size buffer for token information
		GetTokenInformation(hToken, TokenUser, NULL, 0, &dwTuBuff);
		if (dwTuBuff == 0)
		{
			dwError = ::GetLastError(); 
			throw _T("CRegKeyTools::GetUserSID [OpenProcessToken(NULL)]");
		}

		// Allocate memory for token information
		if ((pTuInfo = (PTOKEN_USER)::new BYTE[(int)dwTuBuff]) == NULL)
		{
			dwError = ::GetLastError(); 
			throw _T("CRegKeyTools::GetUserSID [Allocate memory ]");
		}

		// Obtain token information
		memset((LPVOID)pTuInfo, 0, dwTuBuff);
		if (!GetTokenInformation(hToken, TokenUser, pTuInfo,
            dwTuBuff, &dwTuBuff))
		{
			dwError = ::GetLastError(); 
			throw _T("CRegKeyTools::GetUserSID [GetTokenInformation()]");
		}

		// ConvertSidToStringSid
		if (!GetTextualSid(pTuInfo->User.Sid, m_szUserSID, &dwSIDLen))
		{
			dwError = ::GetLastError(); 
			throw _T("CRegKeyTools::GetUserSID [GetTextualSid()]");
		}
		bRetVal = TRUE;
	}
	catch(LPCTSTR lpszErr)
	{
		if (bIsSupport)
			ShowErrorMsg((HRESULT)dwError, lpszErr);
	}

    if (pTuInfo != NULL)
	{
		::delete [] pTuInfo;
		pTuInfo = NULL;
	}
	if (hToken != NULL)
	{
		::CloseHandle(hToken);
		hToken = NULL;
	}
	if (hProcess != NULL)
	{
		::CloseHandle(hProcess);
		hProcess = NULL;
	}
	return bRetVal;
}

BOOL CRegKeyTools::GetTextualSid(PSID pSid, LPTSTR szTextualSid, 
      LPDWORD dwBufferLen)
{
   PSID_IDENTIFIER_AUTHORITY psia;
   DWORD dwSubAuthorities;
   DWORD dwSidRev = SID_REVISION;
   DWORD dwCounter;
   DWORD dwSidSize;

   // Test if SID passed in is valid.
   if(!IsValidSid(pSid)) 
      return FALSE;

   // Obtain SidIdentifierAuthority.
   psia = GetSidIdentifierAuthority(pSid);

   // Obtain sidsubauthority count.
   dwSubAuthorities = *GetSidSubAuthorityCount(pSid);

   // Compute buffer length.
   // S-SID_REVISION- + identifierauthority- + subauthorities- + NULL
   dwSidSize = (15 + 12 + (12 * dwSubAuthorities) + 1) * sizeof(_TCHAR);

   // Check provided buffer length.
   // If not large enough, indicate proper size and setlasterror
   if (*dwBufferLen < dwSidSize)
   {
     *dwBufferLen = dwSidSize;
     SetLastError(ERROR_INSUFFICIENT_BUFFER);
     return FALSE;
   }
   size_t nBufferLen = dwSidSize;

   // Prepare S-SID_REVISION-.
#if _SECURE_ATL
   dwSidSize = sprintf_s(szTextualSid, nBufferLen, TEXT("S-%lu-"), dwSidRev);
#else
   dwSidSize = _stprintf(szTextualSid, TEXT("S-%lu-"), dwSidRev);
#endif

   // Prepare SidIdentifierAuthority.
   if ((psia->Value[0] != 0) || (psia->Value[1] != 0))
   {
#if _SECURE_ATL
      dwSidSize += sprintf_s(szTextualSid + lstrlen(szTextualSid),
		    nBufferLen,
            TEXT("0x%02hx%02hx%02hx%02hx%02hx%02hx"),
            (USHORT) psia->Value[0],
            (USHORT) psia->Value[1],
            (USHORT) psia->Value[2],
            (USHORT) psia->Value[3],
            (USHORT) psia->Value[4],
            (USHORT) psia->Value[5]);
#else
      dwSidSize += _stprintf(szTextualSid + lstrlen(szTextualSid),
            TEXT("0x%02hx%02hx%02hx%02hx%02hx%02hx"),
            (USHORT) psia->Value[0],
            (USHORT) psia->Value[1],
            (USHORT) psia->Value[2],
            (USHORT) psia->Value[3],
            (USHORT) psia->Value[4],
            (USHORT) psia->Value[5]);
#endif
   }
   else
   {
#if _SECURE_ATL
	   dwSidSize += sprintf_s(szTextualSid + lstrlen(szTextualSid),
		    nBufferLen,
            TEXT("%lu"),
            (ULONG) (psia->Value[5]      ) +
            (ULONG) (psia->Value[4] <<  8) +
            (ULONG) (psia->Value[3] << 16) +
            (ULONG) (psia->Value[2] << 24));
#else
	   dwSidSize += _stprintf(szTextualSid + lstrlen(szTextualSid),
            TEXT("%lu"),
            (ULONG) (psia->Value[5]      ) +
            (ULONG) (psia->Value[4] <<  8) +
            (ULONG) (psia->Value[3] << 16) +
            (ULONG) (psia->Value[2] << 24));
#endif
   }

   // Loop through SidSubAuthorities.
   for (dwCounter = 0; dwCounter < dwSubAuthorities; dwCounter++)
   {
#if _SECURE_ATL
      dwSidSize += sprintf_s(szTextualSid + dwSidSize, nBufferLen, TEXT("-%lu"),
            *GetSidSubAuthority(pSid, dwCounter));
#else
      dwSidSize += _stprintf(szTextualSid + dwSidSize, TEXT("-%lu"),
            *GetSidSubAuthority(pSid, dwCounter));
#endif
   }

   return TRUE;
}

